import 'package:collection/collection.dart';
import 'package:device_info_plus/device_info_plus.dart';
import 'package:image_picker/image_picker.dart';

import '../../config.dart';
import '../../logging.dart';
import '../permission/permission_result.dart';
import '../permission/permission_service.dart';
import '../permission/permission_type.dart';
import 'image_source_type.dart';
import 'image_types.dart';

abstract class ImageService {
  Future<List<String>?> pickImages({double? maxWidth, double? maxHeight});
  Future<String?> pickImage({ImageSourceType? imageSourceType, double? maxWidth, double? maxHeight});
}

class ImageServiceImpl extends ImageService {
  static const String _movePathSegment = "moveca";
  static const String _saiMobileAppsHost = "media.saimobileapps.in";
  static const String _ebayHost = "ebayimg";
  static const String _ebayImageKey1 = "\$_";
  static const String _ebayImageKey2 = "s-l";

  final int imageQuality = 100;
  final double maxWidth = 800;
  final double maxHeight = 800;
  final ImagePicker picker;
  final Tracker tracker;

  final ImageSource imageSource = ImageSource.gallery;
  PermissionService permissionService;
  ConfigRepository configRepository;
  DeviceInfoPlugin deviceInfoPlugin;
  DevicePlatform platform;

  final Map<String, List<int>> imageSizes = {};

  ImageServiceImpl({
    required this.permissionService,
    required this.deviceInfoPlugin,
    required this.platform,
    required this.picker,
    required this.configRepository,
    required this.tracker,
  });

  @override
  Future<List<String>?> pickImages({double? maxWidth, double? maxHeight}) async {
    final permissionStatus = await permissionService.checkPermission(PermissionType.photos);
    if (permissionStatus == PermissionResult.denied || permissionStatus == PermissionResult.permanentlyDenied) return null;

    if (platform.isIOS) {
      //Image_picker plugin iOS restrictions:
      //- Multiple images works only from iOS 14 and above.
      //- Impossible to pick HEIC images on the iOS simulator in iOS 14+
      IosDeviceInfo iosDeviceInfo = await deviceInfoPlugin.iosInfo;
      if (iosDeviceInfo.systemVersion != null) {
        final double? version = double.tryParse(iosDeviceInfo.systemVersion!);
        if (version != null && version < 14) {
          final image = await pickImage(
            imageSourceType: ImageSourceType.gallery,
            maxWidth: maxWidth ?? this.maxWidth,
            maxHeight: maxHeight ?? this.maxHeight,
          );
          return image != null ? [image] : null;
        }
      }
    }
    List<XFile>? images = await picker.pickMultiImage(
      maxWidth: maxWidth ?? this.maxWidth,
      maxHeight: maxHeight ?? this.maxHeight,
      imageQuality: imageQuality,
    );
    return images?.map((image) => image.path).toList();
  }

  @override
  Future<String?> pickImage({ImageSourceType? imageSourceType, double? maxWidth, double? maxHeight}) async {
    final permissionStatus = await permissionService.checkPermission(PermissionType.photos);
    if (permissionStatus == PermissionResult.denied || permissionStatus == PermissionResult.permanentlyDenied) return null;
    XFile? image = await picker.pickImage(
      source: imageSourceType?.imageSource ?? imageSource,
      maxWidth: maxWidth ?? this.maxWidth,
      maxHeight: maxHeight ?? this.maxHeight,
      imageQuality: imageQuality,
    );
    return image?.path;
  }

  int _findClosestSize({required int width, required List<int> sizes}) {
    var closest = sizes.reduce((prev, curr) {
      return ((curr - width).abs() < (prev - width).abs() ? curr : prev);
    });
    return closest;
  }

  /// Temporary until we no longer have a dependency on EPS.
  /// Intentionally dumb so it's easier to remove when no longer required
  Uri _ebayOptimizedImage({
    required Uri uri,
    required int width,
    required ImageTypes type,
  }) {
    if (uri.pathSegments.last.contains(_ebayImageKey1)) {
      Map<int, int> sizes = {
        0: 100,
        1: 400,
        12: 500,
        14: 64,
        23: 96,
        26: 200,
        35: 300,
        45: 1200,
        56: 140,
        57: 1600,
        58: 640,
        59: 960,
        62: 225,
        86: 1024,
        118: 275
      };
      var imageWidth = _findClosestSize(
        width: width,
        sizes: sizes.values.toList(),
      );

      var sizeKey = sizes.entries
          .firstWhereOrNull(
            (element) => element.value == imageWidth,
          )
          ?.key;
      if (sizeKey != null) {
        var segments = List<String>.from(uri.pathSegments)..last = "$_ebayImageKey1$sizeKey.JPG";
        return uri.replace(
          pathSegments: segments,
        );
      }

      return uri;
    } else if (uri.pathSegments.last.contains(_ebayImageKey2)) {
      var sizes = [
        48,
        50,
        55,
        56,
        57,
        60,
        64,
        65,
        66,
        70,
        72,
        75,
        80,
        90,
        96,
        99,
        100,
        115,
        122,
        124,
        132,
        135,
        140,
        147,
        148,
        150,
        165,
        175,
        185,
        190,
        198,
        200,
        220,
        221,
        225,
        275,
        280,
        286,
        296,
        300,
        400,
        500,
        640,
        960,
        1024,
        1200,
        1600,
      ];
      var imageWidth = _findClosestSize(width: width, sizes: sizes);
      var segments = List<String>.from(uri.pathSegments)..last = "$_ebayImageKey2$imageWidth.${type.name}";

      return uri.replace(
        pathSegments: segments,
      );
    }
    return uri;
  }
}

class DevicePlatform {
  final bool isAndroid;
  final bool isIOS;

  DevicePlatform({required this.isAndroid, required this.isIOS});
}
