import 'dart:async';

import 'package:geocoding/geocoding.dart';
import 'package:geolocator/geolocator.dart' as geo;

import 'models/location_permission.dart';
import 'models/position.dart';

abstract class LocationService {
  Future<bool> isLocationServicesEnabled();
  Future<PositionResponse> getCurrentPosition({bool requestPermissionAfterLocationCheck = true});
  Future<PositionResponse> getLastKnownPosition({bool requestPermissionAfterLocationCheck = true});
  Future<LocationPermission> checkPermissions();
  Future<List<Placemark>> getPlacemarksFromCoordinates(double latitude, double longitude);
  Future<List<Location>> getLocationFromAddress(String address);
  FutureOr<double> calculateDistance({
    required double startLat,
    required double startLong,
    required double endLat,
    required double endLong,
  });
  Future<LocationPermission> requestPermissions();
  Future<void> openAppSettings();
}

class LocationServiceImpl implements LocationService {
  @override
  FutureOr<double> calculateDistance({
    required double startLat,
    required double startLong,
    required double endLat,
    required double endLong,
  }) {
    return geo.Geolocator.distanceBetween(
      startLat,
      startLong,
      endLat,
      endLong,
    );
  }

  @override
  Future<LocationPermission> checkPermissions() async {
    var permission = await geo.Geolocator.checkPermission();
    return _mapGeoPermissionToPermission(permission: permission);
  }

  Future<FailedPositionResponse?> checkPermissionsForPositionResponse({bool requestPermissionAfterCheck = true}) async {
    bool serviceEnabled;
    LocationPermission permission;

    serviceEnabled = await isLocationServicesEnabled();
    if (!serviceEnabled) {
      return FailedPositionResponse(
        error: PositionError.servicesDisabled,
      );
    }

    permission = await checkPermissions();
    if (permission == LocationPermission.denied && requestPermissionAfterCheck) {
      permission = await requestPermissions();
    }
    if (permission == LocationPermission.denied || permission == LocationPermission.deniedForever) {
      return FailedPositionResponse(
        error: PositionError.permissionDenied,
      );
    }
    return null; //permission allowed
  }

  @override
  Future<PositionResponse> getCurrentPosition({bool requestPermissionAfterLocationCheck = true}) async {
    var failedResponse = await checkPermissionsForPositionResponse(requestPermissionAfterCheck: requestPermissionAfterLocationCheck);
    if (failedResponse != null) return failedResponse;

    var position = await geo.Geolocator.getCurrentPosition();
    return SuccessfulPositionResponse(
      position: Position(
        longitude: position.longitude,
        latitude: position.latitude,
        accuracy: position.accuracy,
        timestamp: position.timestamp,
      ),
    );
  }

  @override
  Future<PositionResponse> getLastKnownPosition({bool requestPermissionAfterLocationCheck = true}) async {
    var failedResponse = await checkPermissionsForPositionResponse(requestPermissionAfterCheck: requestPermissionAfterLocationCheck);
    if (failedResponse != null) return failedResponse;

    var position = await geo.Geolocator.getLastKnownPosition();
    if (position == null) {
      return FailedPositionResponse(error: PositionError.notFound);
    }
    return SuccessfulPositionResponse(
      position: Position(
        longitude: position.longitude,
        latitude: position.latitude,
        accuracy: position.accuracy,
        timestamp: position.timestamp,
      ),
    );
  }

  @override
  Future<bool> isLocationServicesEnabled() async {
    return await geo.Geolocator.isLocationServiceEnabled();
  }

  @override
  Future<LocationPermission> requestPermissions() async {
    var permission = await geo.Geolocator.requestPermission();
    return _mapGeoPermissionToPermission(permission: permission);
  }

  LocationPermission _mapGeoPermissionToPermission({
    required geo.LocationPermission permission,
  }) {
    switch (permission) {
      case geo.LocationPermission.denied:
        return LocationPermission.denied;
      case geo.LocationPermission.deniedForever:
        return LocationPermission.deniedForever;
      case geo.LocationPermission.whileInUse:
        return LocationPermission.whileInUse;
      case geo.LocationPermission.always:
        return LocationPermission.always;
      case geo.LocationPermission.unableToDetermine:
        throw Exception("Not supported");
    }
  }

  // coverage:ignore-start
  // Excluded from tests because of dependence on external APIs.
  @override
  Future<List<Placemark>> getPlacemarksFromCoordinates(double latitude, double longitude) async {
    return await placemarkFromCoordinates(latitude, longitude);
  }

  // Excluded from tests because of dependence on external APIs.
  @override
  Future<List<Location>> getLocationFromAddress(String address) async {
    return await locationFromAddress(address);
  }

  // Excluded from tests because of interaction with native code to open the Settings app.
  @override
  Future<void> openAppSettings() async {
    geo.Geolocator.openAppSettings();
  }
  // coverage:ignore-end
}
