import 'dart:async';
import 'dart:io';

import 'package:dio/dio.dart';
import 'package:uuid/uuid.dart';

import '../../../auth.dart';
import '../../../config.dart';
import '../../../models/auth_token.dart';
import '../../../network.dart';
import '../models/refresh_token_payload_model.dart';
import '../models/token_error_messages.dart';

class AuthenticatedInterceptor extends QueuedInterceptorsWrapper {
  final AuthRepository authRepository;
  final NetworkClient refreshTokenClient;
  final String uuid = const Uuid().v4();

  late final LoginDelegate loginDelegate = Injector.get();
  final String refreshTokenPath = "/token";

  AuthenticatedInterceptor({
    required this.authRepository,
    required this.refreshTokenClient,
  });

  @override
  void onRequest(RequestOptions options, RequestInterceptorHandler handler) async {
    // options.headers.putIfAbsent(CustomHttpHeaders.platformHeader, () {
    //   if (Platform.isIOS) {
    //     return iosPlatformHeaderValue;
    //   } else {
    //     return androidPlatformHeaderValue;
    //   }
    // });

    var authToken = authRepository.auth?.authToken;
    if (authToken == null) {
      handler.reject(DioError(requestOptions: options, error: "No auth token"));
      return;
    }

    if (authToken.isExpired()) {
      Completer completer = Completer();
      options.cancelToken?.whenCancel.then((error) {
        if (!completer.isCompleted) {
          completer.complete();
        }
      });

      var refreshedToken = await refreshToken(
        refreshToken: authToken.refreshToken,
        onCancel: completer,
      );

      if (refreshedToken == null) {
        var loginResponse = await loginDelegate.loginAsync();
        if (loginResponse == null) {
          handler.reject(DioError(requestOptions: options, error: "Could not refresh token"));
          authRepository.setAuth(null);
          return;
        } else {
          authToken = loginResponse.authToken;
        }
      } else {
        authToken = refreshedToken;
        authRepository.setAuth(refreshedToken);
      }
    }

    var tokenType = authToken.tokenType;
    var accessToken = authToken.accessToken;

    options.headers.addEntries(
      [
        MapEntry(HttpHeaders.authorizationHeader, "${capitalize(tokenType)} $accessToken"),
      ],
    );

    super.onRequest(options, handler);
  }

  @override
  void onError(DioError err, ErrorInterceptorHandler handler) async {
    if (_shouldPromptLogin(err.response)) {
      try {
        var loginResponse = await loginDelegate.loginAsync(forceLogin: true);
        if (loginResponse != null) {
          var retry = await Dio(BaseOptions(baseUrl: err.requestOptions.baseUrl)).request(
            err.requestOptions.path,
            cancelToken: err.requestOptions.cancelToken,
            data: err.requestOptions.data,
            onReceiveProgress: err.requestOptions.onSendProgress,
            onSendProgress: err.requestOptions.onSendProgress,
            queryParameters: err.requestOptions.queryParameters,
            options: Options(
              headers: err.requestOptions.headers
                ..addEntries(
                  [
                    MapEntry(HttpHeaders.authorizationHeader, "${capitalize(loginResponse.authToken.tokenType)} ${loginResponse.authToken.accessToken}"),
                  ],
                ),
            ),
          );

          handler.resolve(retry);
        }
      } on Exception catch (_) {
        super.onError(err, handler);
      }
    } else {
      super.onError(err, handler);
    }
  }

  Future<AuthToken?> refreshToken({
    required String refreshToken,
    required Completer onCancel,
  }) async {
    try {
      var response = await refreshTokenClient.post(
        path: refreshTokenPath,
        data: RefreshTokenPayloadModel(
          refreshToken: refreshToken,
        ),
        onCancel: onCancel,
      );

      return AuthToken(
        accessToken: response.data["access_token"],
        idToken: response.data["id_token"],
        refreshToken: response.data["refresh_token"],
        tokenType: response.data["token_type"],
        scope: response.data["scope"],
        expiryDate: DateTime.now().add(
          Duration(seconds: response.data["expires_in"]),
        ),
      );
    } on NetworkException catch (_) {
      return null;
    } on Exception catch (_) {
      return null;
    }
  }

  String capitalize(String str) => str[0].toUpperCase() + str.substring(1);

  /// Used to determine if Login should be presented during an API error.
  /// Currently based off of CIS error messages related to the JWT token
  bool _shouldPromptLogin(Response? response) {
    if (response == null) {
      return false;
    }
    var data = response.data;
    if (data is Map && data.containsKey(TokenErrorMessages.errorMessageKey)) {
      var message = data[TokenErrorMessages.errorMessageKey];
      if ([
        TokenErrorMessages.jwtNotConfigured,
        TokenErrorMessages.missingJwt,
      ].contains(message)) {
        return true;
      }
    }
    return false;
  }
}
