import 'dart:async';

import 'package:dio/dio.dart';

import '../../logging.dart';
import '../../network.dart';
import 'extensions/network_error_extension.dart';
import 'interceptors/base_interceptor.dart';

abstract class NetworkClient {
  Future<NetworkResponse> get({
    required String path,
    required Completer onCancel,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? extras,
  });
  Future<NetworkResponse> post<T>({
    required String path,
    required Completer onCancel,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? extras,
    NetworkResponseType? responseType,
    T? data,
  });
  Future<NetworkResponse> patch<T>({
    required String path,
    required Completer onCancel,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? extras,
    T? data,
  });
  Future<NetworkResponse> delete<T>({
    required String path,
    required Completer onCancel,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? extras,
    T? data,
  });
  Future<NetworkResponse> put<T>({
    required String path,
    required Completer onCancel,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? extras,
    T? data,
  });
}

class NetworkClientImpl implements NetworkClient {
  static const String _get = "GET";
  static const String _post = "POST";
  static const String _patch = "PATCH";
  static const String _delete = "DELETE";
  static const String _put = "PUT";

  final Dio dio;
  final ConnectivityInfo connectivity;

  NetworkClientImpl({
    required this.dio,
    required ErrorTracker errorTracker,
    required this.connectivity,
  }) {
    dio.interceptors.addAll(
      [
        BaseInterceptor(errorTracker: errorTracker),
      ],
    );
  }

  @override
  Future<NetworkResponse> get({
    required String path,
    required Completer onCancel,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? extras,
  }) async {
    var response = await _request(
      path: path,
      method: _get,
      headers: headers,
      queryParameters: queryParameters,
      extras: extras,
      onCancel: onCancel,
    );

    return NetworkResponse(
      statusCode: response.statusCode,
      data: response.data,
      redirects: response.redirects
          .map(
            (e) => Redirect(
              location: e.location,
              statusCode: e.statusCode,
            ),
          )
          .toList(),
    );
  }

  @override
  Future<NetworkResponse> post<T>({
    required String path,
    required Completer onCancel,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? extras,
    NetworkResponseType? responseType,
    T? data,
  }) async {
    var response = await _request<T>(
      path: path,
      method: _post,
      headers: headers,
      queryParameters: queryParameters,
      extras: extras,
      data: data,
      responseType: responseType,
      onCancel: onCancel,
    );

    return NetworkResponse(
      statusCode: response.statusCode,
      data: response.data,
      redirects: response.redirects
          .map(
            (e) => Redirect(
              location: e.location,
              statusCode: e.statusCode,
            ),
          )
          .toList(),
    );
  }

  @override
  Future<NetworkResponse> patch<T>({
    required String path,
    required Completer onCancel,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? extras,
    NetworkResponseType? responseType,
    T? data,
  }) async {
    var response = await _request<T>(
      path: path,
      method: _patch,
      headers: headers,
      queryParameters: queryParameters,
      extras: extras,
      data: data,
      responseType: responseType,
      onCancel: onCancel,
    );

    return NetworkResponse(
      statusCode: response.statusCode,
      data: response.data,
      redirects: response.redirects
          .map(
            (e) => Redirect(
              location: e.location,
              statusCode: e.statusCode,
            ),
          )
          .toList(),
    );
  }

  @override
  Future<NetworkResponse> delete<T>({
    required String path,
    required Completer onCancel,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? extras,
    NetworkResponseType? responseType,
    T? data,
  }) async {
    var response = await _request<T>(
      path: path,
      method: _delete,
      headers: headers,
      queryParameters: queryParameters,
      extras: extras,
      data: data,
      responseType: responseType,
      onCancel: onCancel,
    );

    return NetworkResponse(
      statusCode: response.statusCode,
      data: response.data,
      redirects: response.redirects
          .map(
            (e) => Redirect(
              location: e.location,
              statusCode: e.statusCode,
            ),
          )
          .toList(),
    );
  }

  @override
  Future<NetworkResponse> put<T>({
    required String path,
    required Completer onCancel,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? extras,
    NetworkResponseType? responseType,
    T? data,
  }) async {
    var response = await _request<T>(
      path: path,
      method: _put,
      headers: headers,
      queryParameters: queryParameters,
      extras: extras,
      data: data,
      responseType: responseType,
      onCancel: onCancel,
    );

    return NetworkResponse(
      statusCode: response.statusCode,
      data: response.data,
      redirects: response.redirects
          .map(
            (e) => Redirect(
              location: e.location,
              statusCode: e.statusCode,
            ),
          )
          .toList(),
    );
  }

  Future<Response> _request<T>({
    required String method,
    required String path,
    required Completer onCancel,
    Map<String, dynamic>? headers,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? extras,
    NetworkResponseType? responseType,
    T? data,
  }) async {
    try {
      if (!(await connectivity.isConnected())) {
        throw NetworkException(
          error: NetworkError.connection,
          statusCode: null,
        );
      }
      CancelToken cancelToken = CancelToken();
      onCancel.future.whenComplete(() => cancelToken.cancel());
      return await dio.request(
        path,
        data: await _transformData(data),
        options: Options(
          method: method.toUpperCase(),
          headers: headers,
          extra: extras,
          responseType: responseType?.getType,
        ),
        queryParameters: queryParameters,
        cancelToken: cancelToken,
      );
    } on DioError catch (ex) {
      throw NetworkException(
        error: ex.type.toNetworkError(),
        statusCode: ex.response?.statusCode,
        response: ex.response?.data,
      );
    }
  }

  Future<dynamic> _transformData(data) async {
    if (data is MultipartFileFormData) {
      var file = await MultipartFile.fromFile(data.filePath);
      return FormData.fromMap(
        data.values
          ..addEntries(
            [
              MapEntry(
                "file",
                file,
              ),
            ],
          ),
      );
    }
    //No transformation necessary
    return data;
  }
}
